@@ -43,10 +43,10 @@
 #include <time.h>
 #include <netinet/in.h>
 #include <assert.h>
-#include <sys/types.h>
-#include <sys/socket.h>
-
 #include <linux/types.h>
+#include <sys/socket.h>
+#include <sys/uio.h>
+
 #include <linux/netlink.h>
 
 #include <libnfnetlink/libnfnetlink.h>
@@ -85,6 +85,7 @@ struct nfnl_handle {
 	u_int32_t		subscriptions;
 	u_int32_t		seq;
 	u_int32_t		dump;
+	u_int32_t		rcv_buffer_size;	/* for nfnl_catch */
 	struct nlmsghdr 	*last_nlhdr;
 	struct nfnl_subsys_handle subsys[NFNL_MAX_SUBSYS+1];
 };
@@ -184,6 +185,7 @@ struct nfnl_handle *nfnl_open(void)
 		goto err_close;
 	}
 	nfnlh->seq = time(NULL);
+	nfnlh->rcv_buffer_size = NFNL_BUFFSIZE;
 
 	/* don't set pid here, only first socket of process has real pid !!! 
 	 * binding to pid '0' will default */
@@ -210,6 +212,19 @@ err_free:
 	return NULL;
 }
 
+/**
+ * nfnl_set_rcv_buffer_size - set the size of the receive buffer
+ * @h: libnfnetlink handler
+ * @size: buffer size
+ *
+ * This function sets the size of the receive buffer size, i.e. the size
+ * of the buffer used by nfnl_recv. Default value is 4096 bytes.
+ */
+void nfnl_set_rcv_buffer_size(struct nfnl_handle *h, unsigned int size)
+{
+	h->rcv_buffer_size = size;
+}
+
 /**
  * nfnl_subsys_open - open a netlink subsystem
  * @nfnlh: libnfnetlink handle
@@ -513,7 +528,7 @@ int nfnl_listen(struct nfnl_handle *nfnlh,
 			       void *), void *jarg)
 {
 	struct sockaddr_nl nladdr;
-	char buf[NFNL_BUFFSIZE];
+	char buf[NFNL_BUFFSIZE] __attribute__ ((aligned));
 	struct iovec iov;
 	int remain;
 	struct nlmsghdr *h;
@@ -622,7 +637,7 @@ int nfnl_talk(struct nfnl_handle *nfnlh, struct nlmsghdr *n, pid_t peer,
 	      int (*junk)(struct sockaddr_nl *, struct nlmsghdr *n, void *),
 	      void *jarg)
 {
-	char buf[NFNL_BUFFSIZE];
+	char buf[NFNL_BUFFSIZE] __attribute__ ((aligned));
 	struct sockaddr_nl nladdr;
 	struct nlmsghdr *h;
 	unsigned int seq;
@@ -784,16 +799,16 @@ int nfnl_nfa_addattr_l(struct nfattr *nfa, int maxlen, int type,
 	assert(maxlen > 0);
 	assert(type >= 0);
 
-	if ((NFA_OK(nfa, nfa->nfa_len) + len) > maxlen) {
+	if (NFA_ALIGN(nfa->nfa_len) + len > maxlen) {
 		errno = ENOSPC;
 		return -1;
 	}
 
-	subnfa = (struct nfattr *)(((char *)nfa) + NFA_OK(nfa, nfa->nfa_len));
+	subnfa = (struct nfattr *)(((char *)nfa) + NFA_ALIGN(nfa->nfa_len));
 	subnfa->nfa_type = type;
 	subnfa->nfa_len = len;
 	memcpy(NFA_DATA(subnfa), data, alen);
-	nfa->nfa_len = (NLMSG_ALIGN(nfa->nfa_len) + len);
+	nfa->nfa_len = NFA_ALIGN(nfa->nfa_len) + len;
 
 	return 0;
 }
@@ -943,7 +958,7 @@ void nfnl_build_nfa_iovec(struct iovec *iov, struct nfattr *nfa,
  *
  * This function returns the new size of the socket buffer.
  */
-unsigned int nfnl_rcvbufsiz(struct nfnl_handle *h, unsigned int size)
+unsigned int nfnl_rcvbufsiz(const struct nfnl_handle *h, unsigned int size)
 {
 	int status;
 	socklen_t socklen = sizeof(size);
@@ -1129,6 +1144,7 @@ int nfnl_check_attributes(const struct nfnl_handle *h,
 					 * the kernel which we don't understand
 					 * yet. We have to silently ignore this
 					 * for the sake of future compatibility */
+					attr = NFA_NEXT(attr, attrlen);
 					continue;
 				}
 				nfa[flavor - 1] = attr;
@@ -1190,6 +1206,7 @@ int nfnl_handle_packet(struct nfnl_handle *h, char *buf, int len)
 			return -1;
 
 		len -= rlen;
+		buf += rlen;
 	}
 	return 0;
 }
@@ -1453,52 +1470,12 @@ int nfnl_iterator_next(const struct nfnl_handle *h, struct nfnl_iterator *it)
 int nfnl_catch(struct nfnl_handle *h)
 {
 	int ret;
-	unsigned int size = NFNL_BUFFSIZE;
 
 	assert(h);
 
-	/*
-	 * Since nfqueue can send big packets, we don't know how big
-	 * must be the buffer that have to store the received data.
-	 */
-	{
-		unsigned char buf[size];
-		struct sockaddr_nl peer;
-		struct iovec iov = {
-			.iov_len = size,
-		};
-		struct msghdr msg = {
-			.msg_name = (void *) &peer,
-			.msg_namelen = sizeof(peer),
-			.msg_iov = &iov,
-			.msg_iovlen = 1,
-			.msg_control = NULL,
-			.msg_controllen = 0,
-			.msg_flags = 0
-		};
-
-		memset(&peer, 0, sizeof(peer));
-		peer.nl_family = AF_NETLINK;
-		iov.iov_base = buf;
-		iov.iov_len = size;
-
-retry:		ret = recvmsg(h->fd, &msg, MSG_PEEK);
-		if (ret == -1) {
-			/* interrupted syscall must retry */
-			if (errno == EINTR)
-				goto retry;
-			/* otherwise give up */
-			return -1;
-		}
-
-		if (msg.msg_flags & MSG_TRUNC)
-			/* maximum size of data received from netlink */
-			size = 65535;
-	}
-
-	/* now, receive data from netlink */
 	while (1) {
-		unsigned char buf[size];
+		unsigned char buf[h->rcv_buffer_size]
+			__attribute__ ((aligned));
 
 		ret = nfnl_recv(h, buf, sizeof(buf));
 		if (ret == -1) {
